program cfd

  implicit none

  double precision :: gettime

  double precision, allocatable ::  psi(:,:), vel(:,:,:)
  double precision, allocatable ::  psitmp(:,:)

  integer, allocatable :: rgb(:,:,:)

  integer :: scalefactor,  numiter

  integer, parameter :: maxline = 32
  character(len=maxline) :: tmparg

!  Basic sizes of simulation

  integer, parameter :: bbase = 10
  integer, parameter :: hbase = 15
  integer, parameter :: wbase =  5
  integer, parameter :: mbase = 32
  integer, parameter :: nbase = 32

  double precision :: re

!  re = 3.7 seems to be stability limit with Jacobi

!  Some auxiliary parameters and variables

  integer :: m, n, ln, b, h, w
  integer :: iter, i, j

  double precision :: tstart, tstop, ttot, titer, modvsq, hue

!  Read in parameters

  if (command_argument_count() /= 2 .and. command_argument_count() /= 3) then

     write(*,*) 'Usage: cfd <scale> <numiter> [reynolds]'
     stop

  end if

  call get_command_argument(1, tmparg)
  read(tmparg,*) scalefactor
  call get_command_argument(2, tmparg)
  read(tmparg,*) numiter

  re = -1.0

  write(*,fmt='('' Scale factor = '',i3,'', iterations = '', i6)') &
       scalefactor, numiter

!  Calculate b, h & w and m & n
        
  b = bbase*scalefactor 
  h = hbase*scalefactor
  w = wbase*scalefactor 
  m = mbase*scalefactor
  n = nbase*scalefactor

  re = re / dble(scalefactor)

  write(*,fmt='('' Running CFD on '', i4, '' x '', i4, &
               &'' grid using serial program'')') m, n

!  Allocate arrays, including halos on psi and psitmp

  allocate(psi(0:m+1, 0:n+1))

  allocate(psitmp(0:m+1, 0:n+1))


  allocate(vel(2,m,n))
  allocate(rgb(3,m,n))

!  Zero the psi array

  psi(:,:) = 0.0

!  Set the psi boundary condtions which are constant

   call boundarypsi(psi, m, n, b, h, w)


!  Begin iterative Jacobi loop

  write(*,*)
  write(*,*) 'Starting main loop ...'
  write(*,*)
       
  tstart = gettime()

  do iter = 1, numiter

!  Compute the new psi based on the old one

     psitmp(1:m, 1:n) = 0.25d0*(psi(2:m+1, 1:n) + psi(0:m-1, 1:n) + &
          psi(1:m, 2:n+1) + psi(1:m, 0:n-1)     )
     

!  Copy back

     psi(1:m, 1:n) = psitmp(1:m, 1:n)
        
!  End iterative Jacobi loop

     if (mod(iter,1000) == 0) then
        write(*,*) 'completed iteration ', iter
     end if

  end do

  tstop = gettime()

  ttot  = tstop-tstart
  titer = ttot/dble(numiter)

  write(*,*) 
  write(*,*) '... finished'
  write(*,*)
  write(*,fmt='('' Time for '', i6, '' iterations was '',&
       &g11.4, '' seconds'')') numiter, ttot
  write(*,fmt='('' Each individual iteration took '', g11.4, '' seconds'')') &
       titer
  write(*,*)
  write(*,*) 'Writing output file ...'

  do i = 1, m
     do j = 1, n

        vel(1,i,j) =   (psi(i,j+1)-psi(i,j-1)) / 2.0
        vel(2,i,j) = - (psi(i+1,j)-psi(i-1,j)) / 2.0

        modvsq = vel(1,i,j)**2 + vel(2,i,j)**2
        hue = modvsq**0.4

        call hue2rgb(hue, rgb(1,i,j), rgb(2,i,j), rgb(3,i,j))

     end do
  end do

!  Output result

  call writedatafiles(rgb, vel, m, n, scalefactor)

  call writeplotfile(m, n, scalefactor)

! Finish

  write(*,*) ' ... finished'
  write(*,*)
  write(*,*) 'CFD completed'
  write(*,*)

end program cfd


subroutine boundarypsi(psi, m, n, b, h, w)

  implicit none

  integer :: m, n, b, h, w
  
  double precision, dimension(0:m+1, 0:n+1) :: psi

  integer :: i, j

  do i = b+1, b+w-1
     psi(i, 0) = float(i-b)
  end do

  do i = b+w, m
     psi(i, 0) = float(w)
  end do

!  Set the boundary conditions on the right hand side

  do j = 1, h

     psi(m+1,j) = float(w)

  end do

  do j = h+1, h+w-1

     psi(m+1,j) = float(w-j+h)

  end do

end subroutine boundarypsi



subroutine writedatafiles(rgb, vel, m, n, scale)

  implicit none

  integer :: scale, m, n

  integer, dimension(3,m,n) :: rgb(3,m,n)
  double precision, dimension(2,m,n) :: vel

  integer :: i, j, k

  integer, parameter :: iounitvel = 10, iounitcol = 11

!  Open files

  open(unit=iounitcol, file='colourmap.dat', form='formatted')
  open(unit=iounitvel, file='velocity.dat',  form='formatted')

!  Write out

  do j = 1, n
     do i = 1, m

!  Write colour map of velocity magnitude at every point
        
        write(iounitcol,fmt='(i4,1x,i4,1x,i3,1x,i3,1x,i3)') &
             i, j, rgb(1,i,j), rgb(2,i,j), rgb(3,i,j)

!  Only write velocity vectors every "scale" points

        if (mod(i-1,scale) == (scale-1)/2 .and. &
            mod(j-1,scale) == (scale-1)/2         ) then

           write(iounitvel,fmt='(i4,1x,i4,1x,g12.5,1x,g12.5)') &
                i, j, vel(1,i,j), vel(2,i,j)
        end if
             
     end do
  end do

  close(unit=iounitcol)
  close(unit=iounitvel)

end subroutine writedatafiles

subroutine writeplotfile(m, n, scale)

  implicit none

  integer :: m, n, scale
  integer, parameter :: iounit = 10

  open(unit=iounit, file='cfd.plt', form='formatted')

  write(iounit,*) 'set size square'
  write(iounit,*) 'set key off'
  write(iounit,*) 'unset xtics'
  write(iounit,*) 'unset ytics'

  write(iounit,fmt='('' set xrange ['',i4,'':'',i4, '']'')') 1-scale, m+scale
  write(iounit,fmt='('' set yrange ['',i4,'':'',i4, '']'')') 1-scale, n+scale

  write(iounit,fmt='('' plot "colourmap.dat" w rgbimage, "velocity.dat" u 1:2:&
       &('',i2,''*0.75*$3/sqrt($3**2+$4**2)):&
       &('',i2,''*0.75*$4/sqrt($3**2+$4**2)) &
       &with vectors  lc rgb "#7F7F7F"'')') scale, scale

  close(unit=iounit)

end subroutine writeplotfile


subroutine hue2rgb(hue, r, g, b)

  implicit none
  
  double precision :: hue, colfunc

  integer :: r, g, b
  integer, parameter :: rgbmax = 255

  r = rgbmax*colfunc(hue-1.0)
  g = rgbmax*colfunc(hue-0.5)
  b = rgbmax*colfunc(hue    )

end subroutine hue2rgb


double precision function colfunc(x)

  implicit none

  double precision :: x, absx, val

  double precision, parameter :: x1 = 0.2, x2 = 0.5

  absx = abs(x)

  if (absx .gt. x2) then
     val = 0.0
  else if (absx .lt. x1) then
     val = 1.0
  else
     val = 1.0 - ((absx-x1)/(x2-x1))**2
  end if

  colfunc = val
      
end function colfunc


double precision function gettime()

  logical, save :: firstcall = .true.

  integer, parameter :: int32kind = selected_int_kind( 9)
  integer, parameter :: int64kind = selected_int_kind(18)

  integer, parameter :: intkind = int64kind

  integer(kind = intkind) :: count,rate

  double precision, save :: ticktime

  if (firstcall) then

     firstcall = .false.

     call system_clock(count, rate)

     ticktime = 1.0d0/dble(rate)
     gettime  = dble(count)*ticktime

  else

     call system_clock(count)

     gettime = dble(count)*ticktime

  end if

end function gettime
